/****************************************************************************/
/*! \file		matlab.c
	\brief		Visualization in MatLab
	\author		Iliya Voronov, http://powerdsp.narod.ru
	\version	1.2
  
  Visualisation in MatLab environment:
  plot drowing, eye diagram, constellation diagram
*/
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "matlab.h"
#include "engine.h"  //!< MatLab engine


#ifdef _MATLAB_VIS_



/******************************************************************************
Internal data
******************************************************************************/

//! Variable name size
#define MATLAB_NAME_SZ 64

//! Number of graphs
#define MATLAB_PLOT_NUM 4

//! String size
#define MATLAB_STR_SZ (MATLAB_NAME_SZ*MATLAB_PLOT_NUM)

//! Matlab array size
#define MATLAB_ARR_SZ 512

//! Drawing color number
#define MATLAB_COLOR_NUM 7

//! Dot per eye diagram
#define MATLAB_EYE_LEN 4

//! Number of eye traces in array
#define MATLAB_EYE_NUM 8


//! Number points on constellation
#define MATLAB_CST_LEN 16



//! Drawing color
const char * MATLAB_pColor = "bgrcmyk";



//! Graph
typedef struct tagMATLAB_plot {
	mxArray		  * pMlArr;					//!< Pointer to MatLab array object
	double		  * pMlVal;					//!< Pointer to value of MatLab array
	int				ind;					//!< Index in MatLab array
	char			aName[MATLAB_NAME_SZ];	//!< MatLab variable name
//	char        aFigure[MATLAB_STR_SZ]; //!< MatLab figure command
//	char        aTitle [MATLAB_STR_SZ]; //!< MatLab title command
} tMATLAB_plot;

typedef struct tagMATLAB_inst {
	Engine		  * pMl;						//!< MatLab engine handle
	tMATLAB_plot	aPlot[MATLAB_PLOT_NUM];		//!< plot
	int				updCntr;					//!< added element counter before update
	int				updInt;						//!< update interval in elements 
	char			aPlotCmd[MATLAB_STR_SZ];	//!< plot command
	char			aPlegCmd[MATLAB_STR_SZ];	//!< plot legend command
} tMATLAB_inst;


static tMATLAB_inst matlab = { NULL, };

/******************************************************************************
Forward and inline declarations
******************************************************************************/

//! Update commands after adding new variable
void MATLAB_updCmd( tMATLAB_inst * this );


/******************************************************************************
External functions
******************************************************************************/

//! Open MATLAB
/*!	Open MATLAB and create arrays and plots for plotArr function. 
	Should be called before all other functions.
	\param	updInt		plot update interval
*/
void openMatlab( int updInt )
{
	tMATLAB_inst	* this = &matlab;
	int				i;
	tMATLAB_plot	* pPlot;

	memset( this, 0, sizeof(*this) );
	this->updInt = updInt;

	this->pMl = engOpen("\0"); // Connect to MatLab on local machine
	if( this->pMl == NULL ) { // MatLab is not connected
		printf( "Can not establish connection to MatLab\n" );
		return;
	}

	for( i = 0, pPlot = this->aPlot; i < MATLAB_PLOT_NUM; i++, pPlot++ ) {
		pPlot->aName[0] = '\0'; // empty graph
/*
		strcpy( pVis->aName, pDescr->pName );
		n = sprintf( pVis->aFigure, "figure(%d);", (int)(i+1)   );
		ASSERT( n < PARSE_NAME_SZ );
		n = sprintf( pVis->aPlot,   "plot(%s,'%c');", pVis->aName, pMlColor[i%VIS_ML_COLOR_NUM] );
		ASSERT( n < PARSE_NAME_SZ );
		n = sprintf( pVis->aTitle,  "title('%s');", pVis->aName );
		ASSERT( n < PARSE_NAME_SZ );
		for( j = 0; j < strlen( pVis->aTitle ); j++ )
		if( pVis->aTitle[j] == '_' ) // '_' meand subscript in matlab
		pVis->aTitle[j] = '-';
*/
		pPlot->pMlArr = mxCreateDoubleMatrix( 1, MATLAB_ARR_SZ, mxREAL);  // Create variable
		if ( pPlot->pMlArr != NULL ) {
			pPlot->pMlVal  = (double *) mxGetPr( pPlot->pMlArr);
			if ( pPlot->pMlVal == NULL )
				printf("Can not get access to MatLab array\n" );
		} else {
			pPlot->pMlVal = NULL;
			printf("Can not create MatLab array\n");
		}
		if( pPlot->pMlVal ) {
			memset( pPlot->pMlVal, 0, MATLAB_ARR_SZ*sizeof(double) );
			engPutVariable( this->pMl, pPlot->aName,  pPlot->pMlArr ); // Update MatLab variable
		}

	}
	engEvalString ( this->pMl, "close all;" );
	engEvalString ( this->pMl, "hplot = figure();" );
}



//! Plot array
/*!	Plot array with plot command.
	All arrays are plotter in the same window, 
	up to 4 arrays can be plotted simultaneously.
	Array are identifieds by its name, if no array with specified name
	exists, empty channel is allocated for the array.
	New samples are putted to circular buffer of 512 samples, then 
	whole buffer is plotted. 
	To plot 10 complex samples from array arr use 2 following commands
	plotArr( "arr_re", &arr[0].re, 10, 2 );
	plotArr( "arr-im", &arr[0].im, 10, 2 );
	\param	pName		name of array
	\param	pArr		poiner to array of 16-bit samples
	\param	len			number of plotted samples
	\param	step		step from sample to semple in pArr array
*/
void plotArr (	const char	  * pName,	const short	  * pArr,
				unsigned short	len,	unsigned short	step	)
{
	tMATLAB_inst	  * this = &matlab;
	char				aName[MATLAB_NAME_SZ];
	int					i;
	tMATLAB_plot	  * pPlot;

	if( strlen( pName ) >= (MATLAB_NAME_SZ-1) ) {
		strncpy( aName, pName, MATLAB_NAME_SZ );
		aName[MATLAB_NAME_SZ-1] = '\0'; // strncpy may remove tailing \0
		pName = aName;
	}

	for( i = 0; i < MATLAB_PLOT_NUM; i++ )
		if( strcmp( pName, this->aPlot[i].aName ) == 0 )
			break;

	if( i >= MATLAB_PLOT_NUM ) { // no variable with such name
		for( i = 0; i < MATLAB_PLOT_NUM; i++ ) // try to create it
			if( this->aPlot[i].aName[0] == 0 )
				break;
		if( i >= MATLAB_PLOT_NUM ) { // all graph are busy
			printf("Can not create graph %s\n", aName);
			return;
		}

		// new variable created
		strcpy( this->aPlot[i].aName, pName );
		this->aPlot[i].ind = 0;
		MATLAB_updCmd( this );
		engEvalString( this->pMl, "figure(hplot);" ); 
		engEvalString( this->pMl, this->aPlotCmd ); // Plot all variables
		engEvalString( this->pMl, this->aPlegCmd  );
	}

	// put new values
	pPlot = &this->aPlot[i];
	for( i = 0; i < len; i++ ) {
		pPlot->pMlVal[pPlot->ind] = (double)pArr[i*step];
		pPlot->ind = ( pPlot->ind + 1 ) & ( MATLAB_ARR_SZ-1 );
	}
	this->updCntr += len;
	if( this->updCntr < this->updInt )
		return;

	// time to update plot
	this->updCntr = 0;

	engPutVariable( this->pMl, pPlot->aName,  pPlot->pMlArr ); // Update MatLab variable
	engEvalString( this->pMl, "figure(hplot);" );
	engEvalString ( this->pMl, this->aPlotCmd ); // Plot all variables
	engEvalString ( this->pMl, this->aPlegCmd  );

//	engEvalString ( this->pMl, pVis->aFigure ); // Select window;"
//	engEvalString ( this->pMl, pVis->aTitle  ); // Title it

}


//! Plot eye diagram
/*! Plot eye diagram of 4 complex samples per symbol
	New 4 complex samples are put to circular buffer of 16 samples, then
	whole buffer is plotted.
	\param	pCarr		pointer to complex array, format [0].re, [0].im, [1].re ... [3].im
*/
void eyeDiag4( const short	  * pCarr )
{
static mxArray		  * pEyeArr;
static double		  * pEyeRe, * pEyeIm;
	int		i;

	if( pEyeArr == NULL ) { // called first time
		pEyeArr = mxCreateDoubleMatrix( MATLAB_EYE_LEN*MATLAB_EYE_NUM+1, 1, mxCOMPLEX);  // Create variable
		if( pEyeArr != NULL ) {
			pEyeRe = (double *) mxGetPr(pEyeArr);
			pEyeIm = (double *) mxGetPi(pEyeArr);
			if( ( pEyeRe == NULL ) || ( pEyeIm == NULL ) )
				printf("Can not get access to MatLab array\n" );
		} else {
			pEyeIm = pEyeRe = NULL;
			printf("Can not create MatLab array\n");
		}
		if( ( pEyeRe ) && ( pEyeRe ) ) {
			memset( pEyeRe, 0, MATLAB_EYE_LEN*MATLAB_EYE_NUM*sizeof(double) );
			memset( pEyeIm, 0, MATLAB_EYE_LEN*MATLAB_EYE_NUM*sizeof(double) );
			engPutVariable( matlab.pMl, "eyedg",  pEyeArr ); // Update MatLab variable
		}
		engEvalString ( matlab.pMl, "eyedgh = eyediagram(eyedg,4,1,2,'-');" );
	}

	memcpy( &pEyeRe[0], &pEyeRe[MATLAB_EYE_LEN],
			((MATLAB_EYE_NUM-1)*MATLAB_EYE_LEN+1)*sizeof(double)	);
	memcpy( &pEyeIm[0], &pEyeIm[MATLAB_EYE_LEN],
			((MATLAB_EYE_NUM-1)*MATLAB_EYE_LEN+1)*sizeof(double)	);
	for( i = 0; i < MATLAB_EYE_LEN; i++ ) {
		pEyeRe[i+(MATLAB_EYE_NUM-1)*MATLAB_EYE_LEN+1] = pCarr[i*2  ];
		pEyeIm[i+(MATLAB_EYE_NUM-1)*MATLAB_EYE_LEN+1] = pCarr[i*2+1];
	}
	engPutVariable( matlab.pMl, "eyedg",  pEyeArr ); // Update MatLab variable
	engEvalString( matlab.pMl, "eyediagram(eyedg,4,1,2,'-',eyedgh);" );
}


//! Plot constellation diagram
/*! Plot constellation diagram
	New complex sample is put to circular buffer of 17 samples, then
	whole buffer is plotted. 
	The newest samples are show with red color, oldest - with blue.
	\param	pComp		pointer to complex value, format re, im
*/
extern void constDiag(	const short	  * pComp ) {

static mxArray		  * pArrRe = NULL, * pArrIm = NULL;
static double		  * pCstRe, * pCstIm;


	if( ( pArrRe == NULL ) || ( pArrIm == NULL ) ) { // called first time
		pArrRe = mxCreateDoubleMatrix( MATLAB_CST_LEN+1, 1, mxREAL);
		pArrIm = mxCreateDoubleMatrix( MATLAB_CST_LEN+1, 1, mxREAL);
		if( ( pArrRe != NULL ) && ( pArrIm != NULL ) ) {
			pCstRe = (double *) mxGetPr(pArrRe);
			pCstIm = (double *) mxGetPr(pArrIm);
			if( ( pCstRe == NULL ) || ( pCstIm == NULL ) )
				printf("Can not get access to MatLab array\n" );
		} else {
			pCstIm = pCstRe = NULL;
			printf("Can not create MatLab array\n");
		}
		if( ( pCstRe ) && ( pCstRe ) ) {
			memset( pCstRe, 0, (MATLAB_CST_LEN+1)*sizeof(double) );
			memset( pCstIm, 0, (MATLAB_CST_LEN+1)*sizeof(double) );
			engPutVariable( matlab.pMl, "cstdgre",  pArrRe );
			engPutVariable( matlab.pMl, "cstdgim",  pArrIm );
		}
		engEvalString( matlab.pMl, "hcst = figure();" );
		engEvalString( matlab.pMl, "scatter(cstdgre,cstdgim,4,1:17);" );
		engEvalString( matlab.pMl, "axis([-15000 15000 -15000 15000]);" );
	}

	memcpy( &pCstRe[0], &pCstRe[1],	MATLAB_CST_LEN*sizeof(double)	);
	pCstRe[MATLAB_CST_LEN] = pComp[0];
	engPutVariable( matlab.pMl, "cstdgre",  pArrRe );
	memcpy( &pCstIm[0], &pCstIm[1],	MATLAB_CST_LEN*sizeof(double)	);
	pCstIm[MATLAB_CST_LEN] = pComp[1];
	engPutVariable( matlab.pMl, "cstdgim",  pArrIm  );
	engEvalString( matlab.pMl, "figure(hcst);" );
	engEvalString( matlab.pMl, "scatter(cstdgre,cstdgim,4,1:17);" );
	engEvalString( matlab.pMl, "axis([-15000 15000 -15000 15000]);" );
}


//! Close MATLAB
/*!	Close MATLAB. Currently this function just update plot and 
	do not close MATLAB 
*/
void closeMatlab()
{
	tMATLAB_inst	  * this = &matlab;
	int					i;
	tMATLAB_plot	  * pPlot;

	engEvalString( this->pMl, "figure(hplot);" );
	engEvalString ( this->pMl, this->aPlotCmd  ); // Plot all variables
	engEvalString ( this->pMl, this->aPlegCmd  );

	for( i = 0, pPlot = this->aPlot; i < MATLAB_PLOT_NUM; i++, pPlot++ ) {
		if ( pPlot->pMlArr != NULL )  // destroy only actually created arrays
			mxDestroyArray( pPlot->pMlArr );
	}
}




/*****************************************************************************
Internal functions
*****************************************************************************/

//! Update commands after adding new variable
void MATLAB_updCmd( tMATLAB_inst * this )
{
	char			  * pStr, * pEnd;
	int					i, len, n;
	tMATLAB_plot	  *	pPlot;

	// plot command
	pStr = &this->aPlotCmd[0];
	pEnd = &this->aPlotCmd[MATLAB_STR_SZ-2]; // reserv space for 1 character
	strcpy( pStr, "plot(" );
	len = strlen( pStr );
	pStr += len;
	for( i = 0, pPlot = this->aPlot; i < MATLAB_PLOT_NUM; i++, pPlot++ ) {
		if( pPlot->aName[0] == '\0' )
			continue;
		n = pEnd - pStr;
		len = _snprintf( pStr, n, "1:%d,%s,'%c-',", (int)MATLAB_ARR_SZ, pPlot->aName, MATLAB_pColor[i] );
		assert( ( len > 0 ) && ( len < n ) );
		pStr += len;
	}
	pStr--; // remove tailing ','
	strcpy( pStr, ");" );

	// legend command
	pStr = &this->aPlegCmd[0];
	pEnd = &this->aPlegCmd[MATLAB_STR_SZ-2]; // reserv space for 1 character
    strcpy( pStr, "legend(" );
	len = strlen( pStr );
	pStr += len;
	for( i = 0, pPlot = this->aPlot; i < MATLAB_PLOT_NUM; i++, pPlot++ ) {
		if( pPlot->aName[0] == '\0' )
			continue;
		n = pEnd - pStr;
		len = _snprintf( pStr, n, "'%s',", pPlot->aName );
		assert( ( len > 0 ) && ( len < n ) );
		pStr += len;
	}
	pStr--; // remove tailing ','
	strcpy( pStr, ");" );
}




#else //#ifdef _MATLAB_VIS_

// Do nothing, plug for compilation on target
void openMatlab( int updInt ) { }
void plotArr(	const char	  * pName,	const short	  * pArr,
				unsigned short	len,	unsigned short	step	){}
void eyeDiag4(	const short	  * pCarr ) { }
void constDiag(	const short	  * pComp ) { }
void closeMatlab() { }


#endif //#ifdef _MATLAB_VIS_
